/*
   program       : various

   $Source: /home/yee/C/RCS/filter.c,v $
   $Author: yee $
   $Date: 1999/02/10 16:39:47 $
   $Revision: 2.18 $

   This file contains the code implementing various 1-D digital filters.
   The output can be an array of either complex or real numbers.

   This code was written as an experiment in the extensive use of #define
   to determine compilation.  This way, the same code can be used to compile
   many similar functions.  Bugs can be fixed in all of the different
   functions simultaneuously.

   The code was also written as an experiment to allow functions to call other
   functions to dynamically determine behavior.

   Since the code uses a skeleton to generate even filters and calls a
   function to provide actual values, externally defined filters can be used.
   e.g. DefineComplexFilter(W,128,hamming) to put a complex hamming
        filter 128 elements in length into the array W.

   The functions are taken from information derived from Oppenheim & Schafer,
   "Digital Signal Processing", pgs 241-2.  The equations are modified from
   those in Oppenheim & Shafer for
      1) computational efficiency
      2) the equations assume an odd number of elements which is not useful
         in this application.  Thus, "ndim" is substituted for "ndim-1".

   WARNING : not all filters tested.

   Parameters :
      1) #define FILTER_SINGLE|SINGLE for single precision.
         #define FILTER_DOUBLE|DOUBLE for double precision.

      2) #define FILTER_COMPLEX for a complex filter.
         #define FILTER_REAL for a real filter.

      3) #define HAMMING  for a Hamming     filter.
         #define HANNING  for a Hanning     filter.
         #define RECT     for a Rectangular filter.
         #define BARTLETT for a Bartlett    filter.
         #define BLACKMAN for a Blackman    filter.
         #define KAISER   for a Kaiser      filter.
         #define WELCH    for a Welch       filter.

         Note: if Kaiser filter is used, must define the degree of freedom
               in the Kaiser filter.  It is stored in FloatFree[0].

      4) #define FILTER_DEFINE for the setting array to the filter
         #define FILTER_APPEND for additional filtering of an existing
                                   filter or to simply filter data.

      5) #define FILTER_CENTERED if zero-th element of filter is in the
         middle of the array.  
         The default is the zero-th element is the first element.

   Assumptions : none known

   Bugs :
     1) define bessel() in the Kaiser filter. Kaiser filter not working.

   To Do :
     1) this code is TOO clever for maintenance.

   $Log: filter.c,v $
   Revision 2.18  1999/02/10 16:39:47  yee
   *** empty log message ***

   Revision 2.17  1999/01/17 00:02:00  yee
   *** empty log message ***

   Copyright 1997, 1998, 1999 by Conway Yee.  All rights reserved.
 */
#include <math.h>
#include <stdlib.h>

#include "mutil.h"
#include "filter.h"
#include "style.h"

#define INTERNAL_FILTER  /* internal flag to compile FILTER() */
#if defined(HAMMING)
#define FILTER_NAME   hamming
#define FILTER_DEFINITION(ndim,i,free) (0.54 - 0.46*cos(TWO_PI*(i)/(ndim)))
#elif defined(HANNING)
#define FILTER_NAME   hanning
#define FILTER_DEFINITION(ndim,i,free) (0.5*(1.0-cos(TWO_PI*(i)/(ndim))))
#elif defined(RECT)
#define FILTER_NAME   rect
#define FILTER_DEFINITION(ndim,i,free) (1.0)
#elif defined(BARTLETT)
#define FILTER_NAME   bartlett
#define FILTER_DEFINITION(ndim,i,free) (2.0*(i)/(ndim)) 
#elif defined(BLACKMAN)
#define FILTER_NAME   blackman
#define FILTER_DEFINITION(ndim,i,free) (0.42-0.5*cos(TWO_PI*(i)/(ndim))+0.08*cos(4*PI*(i)/(ndim)))
#elif defined(KAISER)
                      /* Note: if Kaiser window function is used, the
                               degree of freedom is stored in FloatFree[0]. */
#define FILTER_NAME   kaiser
#define FILTER_DEFINITION(ndim,i,free) (bessel((*FloatFree)*pow(pow(0.5*(ndim),2.0)-pow((i)-0.5*(ndim),2.0),0.5),0)/bessel((*FloatFree)*0.5*(ndim),0))
#elif defined(WELCH)
#define FILTER_NAME   welch
#define FILTER_DEFINITION(ndim,i,free) (1.0-((i)-0.5*(ndim)/(0.5*(ndim))))
#else
#undef INTERNAL_FILTER  /* externally defined filter */
#endif

#ifdef FILTER_APPEND
#define FILTER_MODE Append
#define OPERATION *=
#elif defined(FILTER_DEFINE)
#define FILTER_MODE Define
#define OPERATION = 
#else
#if !defined(INTERNAL_FILTER)
#error APPEND_VS_DEFINE_NOT_DEFINED
#endif
#endif FILTER_MODE

#ifdef FILTER_REAL
#define FILTER_TYPE Real

#if defined(FILTER_DOUBLE) || defined(DOUBLE)
#define FILTER_ELEMENT double
#elif defined(FILTER_SINGLE) || defined(SINGLE)
#define FILTER_ELEMENT float
#else SINGLE_VS_DOUBLE_NOT_DEFINED
#error SINGLE_VS_DOUBLE_NOT_DEFINED
#endif 

#elif defined(FILTER_COMPLEX)
#define FILTER_TYPE Complex

#if defined(FILTER_DOUBLE) || defined(DOUBLE)
#define FILTER_ELEMENT lcomplex
#elif defined(FILTER_SINGLE) || defined(SINGLE)
#define FILTER_ELEMENT complex
#else SINGLE_VS_DOUBLE_NOT_DEFINED
#error SINGLE_VS_DOUBLE_NOT_DEFINED
#endif 

#else
#if !defined(INTERNAL_FILTER)
#error FILTER_ELEMENT_TYPE_NOT_DEFINED
#endif
#endif FILTER_ELEMENT_TYPE

#ifdef FILTER_CENTERED
#define FILTER_CENTER Centered
#else
#define FILTER_CENTER
#endif

#define CONCAT4(a,b,c,d) a##b##c##d
#define CAT4(a,b,c,d) CONCAT4(a,b,c,d)

#define EVEN_FILTER CAT4(FILTER_MODE,FILTER_TYPE,FILTER_CENTER,Filter)
 
/*****************************************************************************/
/*****************************************************************************/

/* this function provides the skeleton of the generic filter.  The filter is
   assumed to be an even function. 

   If #define FILTER_APPEND, the function is called Append"TYPE"Filter()
   and it superimposes the filter to the existing array W.

   If #define FILTER_DEFINE, the function is called Define"TYPE"Filter()
   and it puts the filter into the array W.

   In both modes, "TYPE" is Real or Complex e.g. AppendRealFilter().

   If #define FILTER_CENTERED, then the zero-th element is in the middle
   and the function name become AppendRealCenterFilter().

   NOTE: The elements are allocated from the first and last elements inward
         toward the middle. 

   WARNING: This code is TOO CLEVER for easy maintenance!!! */

#if !defined(INTERNAL_FILTER)
PUBLIC FILTER_ELEMENT *EVEN_FILTER
(
 FILTER_ELEMENT *W,            /* filter array */
 int adim,                     /* size of array (w/ zero padding) */
 int fdim,                     /* size of filter (w/o zero padding) */
 double (*FilterDefinition)(), /* filter to be used */
 void *free                    /* degrees of freedom */
)
{ /* EVEN_FILTER() */
  int mid;                         /* mid point in filter */
  FILTER_ELEMENT *Wi;              /* filter from first element forward */
  FILTER_ELEMENT *Wf;              /* filter from last element backward */
  int i;                           /* counter */
  FILTER_ELEMENT *Wmid;            /* middle element in array */
  FILTER_ELEMENT *Wfirst;          /* first element in array */
  int dir;                         /* direction that array elements added */
  int first;                       /* first element in filter */
  int count=(fdim>>1)-1;           /* counter */

  if(fdim>adim) return(NULL);  /* error check */

#ifdef FILTER_CENTERED
  {
    int center=adim>>1;
    i=(fdim>>1)-1;
    dir= -1;
    Wmid=W+center;
    Wfirst = W;
    mid = 0;
    first= -1*center;
    Wi= Wmid-(center-1);
    Wf= Wmid+(center-1);
  }
#else !FILTER_CENTERED
  i=1;
  dir = 1;
  Wi= W+1;
  Wf=W+adim-1;
  mid = fdim>>1;
  first = 0;
  Wfirst = W;
  Wmid = W+mid;
#endif CENTERED_FILTER

  for(;count>0;count--,i+=dir,Wi++,Wf--) {
#ifdef FILTER_COMPLEX
      (*Wi).r = (*Wf).r OPERATION FilterDefinition(fdim,i,free);
      (*Wf).i = (*Wi).i = 0.0; 
#elif defined(FILTER_REAL)
      *Wf = *Wi OPERATION FilterDefinition(fdim,i,free);
#else REAL_VS_COMPLEX_NOT_DEFINED
#error
#endif FILTER_REAL|FILTER_COMPLEX
  }

#ifdef FILTER_COMPLEX
    (*Wfirst).r OPERATION FilterDefinition(fdim,first,free);
    (*Wmid).r OPERATION FilterDefinition(fdim,mid,free);
    (*Wfirst).i = (*Wmid).i = 0.0;
#elif defined(FILTER_REAL)
    *Wfirst OPERATION FilterDefinition(fdim,first,free);
    *Wmid OPERATION FilterDefinition(fdim,mid,free);
#else REAL_VS_COMPLEX_NOT_DEFINED
#error
#endif FILTER_REAL|FILTER_COMPLEX

  return(W);
} /* EVEN_FILTER() */
#endif !defined(INTERNAL_FILTER)

/*****************************************************************************/

/* this function turns the macro definition of a given element of a generic
   filter into a function.  It returns a specified element of the generic
   filter.

   This function may not be compiled if only externally defined filters are
   used.*/

#ifdef INTERNAL_FILTER
PUBLIC double FILTER_NAME
(
 int ndim,           /* dimensionality of filter */
 int i,              /* element of filter to return */
 void *free          /* integer degrees of freedom */
)
{ /* FILTER() */
               /* Note: if Kaiser window function is used, the degree of
                        freedom is stored in FloatFree[0]. */
  return(FILTER_DEFINITION(ndim,i,free));
} /* FILTER() */
#endif INTERNAL_FILTER
